//-----------------------------------------------------------------------------
// name: bk-audio.cpp
// desc: audio stuff for bokeh
//
// author: Ge Wang (ge@ccrma.stanford.edu)
//   date: 2013
//-----------------------------------------------------------------------------
#include "bk-audio.h"
#include "bk-globals.h"
#include "bk-sim.h"
#include "y-fft.h"
#include "y-waveform.h"
#include "bk-me.h"
#include <iostream>
using namespace std;



// globals
YFluidSynth * g_synth;
YEcho * g_echo;
double g_now;
double g_nextTime;
int g_prog = 0;



struct Note
{
    int channel;
    float pitch;
    float velocity;
    float duration; // in seconds
    
    Note( int c, float p, float v, float d )
    {
        channel = c;
        pitch = p;
        velocity = v;
        duration = d;
    }
};




vector<Note> g_notes;
int g_noteIndex = 0;
XMutex g_mutex;

// play some notes
void 弹( float 音高, float velocity )
{
    cerr << "音高：" << 音高 << " " << velocity << endl;
    // g_synth->noteOn( 0, 音高, velocity * 128 );
    
    // lock
    g_mutex.acquire();
    // clear notes
    g_notes.clear();
    for( int i = 0; i < 24; i++ )
    {
        g_notes.push_back( Note(0, 音高 + i*2, (1 - i/24.0), .15*(1 - i/24.0) ) );
    }
    // unlock
    g_mutex.release();
    
    // reset the index
    g_noteIndex = 0;
    
    // play now!
    g_nextTime = g_now;
}



//-----------------------------------------------------------------------------
// name: audio_callback
// desc: audio callback
//-----------------------------------------------------------------------------
static void audio_callback( SAMPLE * buffer, unsigned int numFrames, void * userData )
{
    g_now += numFrames;
    
    if( g_now > g_nextTime )
    {
        g_mutex.acquire();
        if( g_noteIndex < g_notes.size() )
        {
            g_synth->noteOn( 0, g_notes[g_noteIndex].pitch, g_notes[g_noteIndex].velocity * 127 );
            g_synth->noteOn( 0, g_notes[g_noteIndex].pitch + 4, g_notes[g_noteIndex].velocity * 127 );
            g_nextTime += g_notes[g_noteIndex].duration * BK_SRATE;
            g_noteIndex++;
        }
        g_mutex.release();
    }
    
    // sum
    SAMPLE sum = 0;
    // num channels
    unsigned int channels = Globals::lastAudioBufferChannels;
    
    // zero out
    memset( Globals::lastAudioBuffer, 0,
           sizeof(SAMPLE)*Globals::lastAudioBufferFrames*channels );
    memset( Globals::lastAudioBufferMono, 0,
           sizeof(SAMPLE)*Globals::lastAudioBufferFrames );
    
    // copy to global buffer
    memcpy( Globals::lastAudioBuffer, buffer,
           sizeof(SAMPLE)*numFrames*channels );
    
    // copy to mono buffer
    for( int i = 0; i < numFrames; i++ )
    {
        // zero out
        sum = 0;
        // loop over channels
        for( int j = 0; j < channels; j++ )
        {
            // sum
            sum += buffer[i*channels + j];
        }
        // set
        Globals::lastAudioBufferMono[i] = sum / channels;
    }
    
    // window it
    for( int i = 0; i < numFrames; i++ )
    {
        // multiply
        Globals::lastAudioBufferMono[i] *= Globals::audioBufferWindow[i];
    }
    // set in the wave
    Globals::waveform->set( Globals::lastAudioBufferMono, numFrames );
    
    // synthesize it
    g_synth->synthesize2( buffer, numFrames );
    // echo it
    // g_echo->synthesize2( buffer, numFrames );
    
    // zero out for output
    // memset( buffer, 0, sizeof(SAMPLE)*numFrames*XAudioIO::numChannels() );
}




//-----------------------------------------------------------------------------
// name: bk_audio_init()
// desc: initialize audio system
//-----------------------------------------------------------------------------
bool bk_audio_init( unsigned int srate, unsigned int frameSize, unsigned channels )
{
    // initialize
    if( !XAudioIO::init( 0, 0, srate, frameSize, channels, audio_callback, NULL ) )
    {
        // done
        return false;
    }
    
    g_synth = new YFluidSynth();
    g_synth->init( srate, 32 );
    g_synth->load( "data/sfonts/rocking8m11e.sf2", "" );
    g_synth->programChange( 0, 47 );
    
    // allocate echo
    g_echo = new YEcho( srate );
    g_echo->setDelay( 0, .25 );
    g_echo->setDelay( 1, .5 );
    
    // allocate
    Globals::lastAudioBuffer = new SAMPLE[frameSize*channels];
    // allocate mono buffer
    Globals::lastAudioBufferMono = new SAMPLE[frameSize];
    // allocate window buffer
    Globals::audioBufferWindow = new SAMPLE[frameSize];
    // set frame size (could have changed in XAudioIO::init())
    Globals::lastAudioBufferFrames = frameSize;
    // set num channels
    Globals::lastAudioBufferChannels = channels;
    
    // compute the window
    hanning( Globals::audioBufferWindow, frameSize );
    
    // create waveform
    Globals::waveform = new YWaveform();
    // place it
    Globals::waveform->loc.y = 1.5f;
    // set the width
    Globals::waveform->setWidth( 2.5f );
    // set the height
    Globals::waveform->setHeight( .75f );
    // initialize it
    Globals::waveform->init( frameSize );
    // active?
    Globals::waveform->active = Globals::renderWaveform;
    
    // add to sim
    Globals::sim->root().addChild( Globals::waveform );
    
    return true;
}




//-----------------------------------------------------------------------------
// name: vq_audio_start()
// desc: start audio system
//-----------------------------------------------------------------------------
bool bk_audio_start()
{
    // start the audio
    if( !XAudioIO::start() )
    {
        // done
        return false;
    }
    
    return true;
}
